#include <stdio.h>
#include <memory.h>
#include <errno.h>
#include "common.h"

static Tcl_Channel		tclStdout, tclStderr;
static Tcl_ChannelType	channelType;

typedef struct string {
	char *ptr;
	int   len;
} string;

static string *get_next_line(char *buf) {
	string *str = NULL;
	int num = 0;
	char *pbuf;

	if (*buf == 0) return NULL;

	str = malloc(sizeof(string));
	str->ptr = malloc(MIRC_MAXSTRING);
	pbuf = str->ptr;

	while (*buf) {
		num++;
		if (*buf == '\r') { buf++; num++; }
		if (*buf == '\n') break;
		*pbuf++ = *buf++;
	}
	*pbuf = 0;
	str->len = num;

	return str->len == 0 ? NULL : str;
}

/* Write Tcl wrapper method */

int Io_Write(ClientData instanceData, const char *buf, int toWrite, int *errorCodePtr) {
	string *pcmd;
	char *sChannel, *pbuf = (char *)buf;
	int start = 0;

	if (*pbuf == 13 || *pbuf == 10) return toWrite;

	if (instanceData == (ClientData)TCL_STDIN ) sChannel = "STDIN";
	if (instanceData == (ClientData)TCL_STDOUT) sChannel = "STDOUT";
	if (instanceData == (ClientData)TCL_STDERR) sChannel = "STDERR";

	CLBCommandf("%s%s_%s_BEGIN", 0, CLB_SIGNAL_PREFIX, MDLL_SIG, sChannel);

	while (pcmd = get_next_line(pbuf)) {
		CLBCommandf("%s%s_%s %s", 0, CLB_SIGNAL_PREFIX, MDLL_SIG, sChannel, pcmd->ptr);
		
		pbuf += pcmd->len;

		free(pcmd->ptr);
		free(pcmd);
	}

	CLBCommandf("%s%s_%s_END", 0, CLB_SIGNAL_PREFIX, MDLL_SIG, sChannel);

	return toWrite;
}

/* Dummy methods */

int Io_Dummy(ClientData instance, ...) { 
	return TCL_OK; 
}

/* Initialization */

int Io_Init(Tcl_Interp *interp) {
	memset(&channelType, 0, sizeof(Tcl_ChannelType));

	channelType.version		  = TCL_CHANNEL_VERSION_2;
	channelType.typeName	  = "virtual";
	channelType.closeProc	  = Io_Dummy;
	channelType.inputProc	  = Io_Dummy;
	channelType.outputProc	  = Io_Write;
	channelType.watchProc	  = Io_Dummy;
	channelType.getHandleProc = Io_Dummy;

	tclStdout = Tcl_CreateChannel(&channelType, "stdout", (ClientData)TCL_STDOUT, TCL_WRITABLE);
	tclStderr = Tcl_CreateChannel(&channelType, "stderr", (ClientData)TCL_STDERR, TCL_WRITABLE);

	/* Register the channels */
	Tcl_RegisterChannel(interp, tclStdout);
	Tcl_RegisterChannel(interp, tclStderr);

	/* Don't do any buffering */
	Tcl_SetChannelOption(interp, tclStdout, "-buffering", "none");
	Tcl_SetChannelOption(interp, tclStderr, "-buffering", "none");

	/*
		Bugfix for stderr channel getting garbage collected in asynchronous calls.
		Reference: http://www.tcl.tk/man/tcl8.4/TclLib/GetStdChan.htm

		This is a workaround for a misfeature in Tcl_SetStdChannel that it fails to 
		do some reference counting housekeeping. This misfeature cannot be corrected 
		without contradicting the assumptions of some existing code that calls 
		Tcl_SetStdChannel. 
	*/
	Tcl_RegisterChannel(NULL, tclStdout);
	Tcl_RegisterChannel(NULL, tclStderr);

	/* Set the channels */
	Tcl_SetStdChannel(tclStdout, TCL_STDOUT);
	Tcl_SetStdChannel(tclStderr, TCL_STDERR);

	return TCL_OK;
}

/* Shutdown / cleanup */

int Io_Shutdown(Tcl_Interp *interp) {
	Tcl_UnregisterChannel(interp, tclStdout);
	Tcl_UnregisterChannel(interp, tclStderr);

	/* Also remove it from global space */
	Tcl_UnregisterChannel(NULL, tclStdout);
	Tcl_UnregisterChannel(NULL, tclStderr);

	return TCL_OK;
}